/*
 * Reads filenames from stdin, copying them
 * to ./filename/MTIME.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <dirent.h>
#ifdef sun
extern char *valloc();
#endif

extern int errno;
char *progname;
int verbose;
int found;	/* # files already present in . (i.e. not created) */
int same;	/* # files same as source file */
int ntrunc;	/* # "found" files needing truncation */
int dir_f=1, dir_b;
struct {
	int total, small, cmp, rm;
} dups;
char *iobuf;
#define IOSIZE (64*1024)

main(argc, argv)
	char **argv;
{
	char *p, buf[MAXPATHLEN*2];
	int c, nfiles;

	progname = argv[0];
	while ((c = getopt(argc, argv, "v")) != EOF)
		switch (c) {
		case 'v':
			verbose++;
			break;
		default:
			break;
		}
	if ((iobuf = valloc(IOSIZE)) == NULL) {
		Perror("valloc", "");
		exit(1);
	}
	umask(07222);
	nfiles = 0;
	while (fgets(buf, sizeof buf, stdin) != NULL) {
		if ((p = strchr(buf, '\n')) != NULL)
			*p = 0;
		if (strlen(buf) >= MAXPATHLEN-16)
			continue;
		if (copyfile(buf) < 0) {
			fprintf(stderr, "%s: premature end at %s\n",
				progname, buf);
			break;
		}
		nfiles++;
	}
	if (verbose) {
		fprintf(stderr, "%s: %d files copied, %d found (%d trunc), %d same, %d dirs made (forward %.1f, back %.1f)\n",
			progname, nfiles, found, ntrunc, same,
			dir_f+dir_b,
			dir_f*100.0/(dir_f+dir_b),
			dir_b*100.0/(dir_f+dir_b));
		fprintf(stderr, "%s: duplicates: %d dirs checked, %d too small, %d files compared, %d dups removed\n",
			progname, dups.total, dups.small, dups.cmp, dups.rm);
	}
	exit(0);
}

copyfile(path)
	char *path;
{
	struct stat st;
	int	i, n, bfd, fd, new;
	char	*cp, *t, bpath[MAXPATHLEN];
	struct timeval tv[2];

	if ((fd = open(path, O_RDONLY, 0)) < 0) {
		/*
		 * File may have been removed under our feet.
		 * Don't make noise about that.
		 */
		if (errno != ENOENT)
			Perror("open", path);
		return 0;
	}
	if (fstat(fd, &st) < 0) {
		Perror("fstat", path);
		close(fd);
		return 0;
	}
	/* Build the file name suffix based on the source file's mtime */
	t = ctime(&st.st_mtime);
	if (t[8] == ' ')
		t[8] = '0';
	for (cp = path; *cp == '/'; cp++)
		;
	sprintf(bpath, "./%s/%c%c%c%c%c-%c%c:%c%c",
		cp, t[4],t[5],t[6],t[8],t[9],t[11],t[12],t[14],t[15]);
	errno = 0;
	if ((bfd = mycreat(bpath, &st, &new)) <= 0) {
		/*
		 * mkdir() can fail with EMLINK when the parent
		 * directory is huge.  Just continue in that case.
		 */
		if (bfd < 0 && errno != EMLINK)
			return -1;
		close(fd);
		return 0;
	}
	if (verbose > 1)
		fprintf(stderr, "%s: cp %s %s\n", progname, path, bpath);
	(void) fchown(bfd, st.st_uid, st.st_gid);
	for (i = 0; (n = read(fd, iobuf, IOSIZE)) > 0; i++)
		if (write(bfd, iobuf, n) != n) {
			Perror("write", bpath);
			(void) close(fd);
			(void) close(bfd);
			(void) unlink(bpath);
			return -1;
		}
	(void) close(fd);
	(void) close(bfd);
	if (i == 0) {		/* nothing copied: lose the inode */
		(void) unlink(bpath);
		return 0;
	}
	/*
	 * Reset the source file's times to what they were,
	 * to avoid smashing the atime every time the file is backed up.
	 * Of course, there's a race condition here.
	 */
	tv[0].tv_sec = st.st_atime;
	tv[1].tv_sec = st.st_mtime;
	tv[0].tv_usec = tv[1].tv_usec = 0;
	(void) utimes(path, tv);
	/* Preserve file times, so "find /backup -newer ..." works */
	(void) utimes(bpath, tv);
	*strrchr(bpath, '/') = 0;
	if (!new)
		rm_dups(bpath);		/* remove duplicates */
	return 0;
}

/*
 *  Create the file s.  Make any intervening directories if needed.
 */
mycreat(s, stp, newp)
char *s;
struct stat *stp;
int *newp;	/* set to true if new dir created */
{
	struct stat sb;
	register char *cp;
	int fd, i;

	/*
	 * The file may already be here, so check the
	 * mtime and size.  In any case, try to use the
	 * same data blocks.  This costs 1 extra fstat,
	 * but avoids copying entire files quite often.
	 */
	if ((fd = open(s, O_CREAT|O_WRONLY, stp->st_mode)) > 0) {
		*newp = 0;
		found++;
		fstat(fd, &sb);
		if (sb.st_mtime == stp->st_mtime &&
		    sb.st_size == stp->st_size) {
			same++;
			close(fd);
			return 0;	/* not an error */
		}
		if (sb.st_size > stp->st_size)
			ftruncate(fd, stp->st_size), ntrunc++;
		return fd;
	}
	*newp = 1;
	/*
	 * Create directories down the path of s in . 
	 * We start from the end of s and work backwards
	 * until mkdir() succeeds, then work forward from
	 * there creating the remaining dirs.
	 * It would be simpler to start at the beginning,
	 * but that costs many more mkdir calls.
	 * Usually only one or 2 calls are needed.
	 */
	cp = s+strlen(s);
	for (;;) {
		while (*--cp != '/')
			if (cp == s)
				return -1;
		*cp = 0;
		i = makedir(s, s+1);
		*cp = '/';
		if (i == 0)
			{dir_b++; break;}
		if (errno != ENOENT) {
			if (errno != EMLINK)
				Perror("mkdir ", s);
			return -1;
		}
	}
	cp++;
	while ((cp = strchr(cp, '/')) != 0) {
		*cp = '\0';
		if (makedir(s, s+1) < 0) {
			if (errno != EMLINK)
				Perror("mkdir ", s);
			return -1;
		}
		dir_f++;
		*cp++ = '/';
	}
	if (cp == s+2) {
		fprintf(stderr, "Gak! on %s\n", s);
		return -1;
	}
	/*
	 * If the file had read, turn on exec in the directory above.
	 */
	cp = strrchr(s, '/');
	*cp = 0;
	if (stat(s+1, &sb) == 0) {
		if (sb.st_mode&S_IREAD)
			sb.st_mode |= S_IEXEC;
		if (sb.st_mode&(S_IREAD>>3))
			sb.st_mode |= S_IEXEC>>3;
		if (sb.st_mode&(S_IREAD>>6))
			sb.st_mode |= S_IEXEC>>6;
		chmod(s, (int)sb.st_mode & ~0222);
	}
	*cp = '/';
	i = creat(s, stp->st_mode);
	if (i < 0)
		Perror("creat", s);
	return i;
}

Perror(op, path)
	char *op, *path;
{
	int e = errno;

	fprintf(stderr, "%s: %s ", progname, op);
	errno = e;
	perror(path);
	errno = e;
}

makedir(path, origpath)
	char *path, *origpath;
{
	struct stat sb;
	int fd;

	if (mkdir(path, 0) < 0) {
		if (errno != EEXIST)
			return -1;
		/*
		 * The path already exists -- if it's
		 * not a directory, remove it and try again.
		 */
		if (stat(path, &sb) < 0)
			return -1;
		if ((sb.st_mode&S_IFMT) != S_IFDIR) {
			(void) unlink(path);
			if (mkdir(path, 0) < 0)
				return -1;
		}
	}
	/* copy stat info from origpath */
	if (stat(origpath, &sb) < 0)
		return -1;
	if ((fd = open(path, 0, 0)) >= 0) {
		(void) fchmod(fd, (int)sb.st_mode & ~0222);
		(void) fchown(fd, sb.st_uid, sb.st_gid);
		(void) close(fd);
	}
	return 0;
}

/*
 * Structure used to sort files in rm_dups.
 */
#define BNAMELEN 11
struct t {
	char name[BNAMELEN+1];	/* known length of /backup file names */
	off_t size;
};

int
comp(a, b)
	struct t *a, *b;
{
	return a->size - b->size;
}

/*
 * Remove redundant copies of files under the given directory.
 */
rm_dups(path)
	char *path;
{
	char file[MAXPATHLEN];
	DIR *dirp;
	struct dirent *dp;
	static struct t *t;
	static int tsize;
	struct t *f1, *f2;
	int nt, low, i;
	struct stat st;

	dups.total++;
	/*
	 * Build a table containing (name,size) pairs for each file.
	 */
	if ((dirp = opendir(path)) == NULL)
		return;
	nt = 0;
	while ((dp = readdir(dirp)) != NULL) {
		if (strlen(dp->d_name) != BNAMELEN)
			continue;
		sprintf(file, "%s/%s", path, dp->d_name);
		if (stat(file, &st))
			continue;
		if (nt >= tsize) {
			if (tsize == 0)
				t = (struct t *) malloc((tsize=8)*sizeof *t);
			else
				t = (struct t *) realloc(t, (tsize *= 2)*sizeof *t);
			if (t == 0)
				Perror("malloc", " "), exit(1);
		}
		strcpy(t[nt].name, dp->d_name);
		t[nt].size = st.st_size;
		nt++;
	}
	closedir(dirp);
	if (nt < 2) {		/* less than 2 files -- nothing to do */
		dups.small++;
		return;
	}
	qsort(t, nt, sizeof *t, comp);	/* sort by size */
	/*
	 * For each group of files of equal size,
	 * compare the first in the group against each of the others,
	 * and remove duplicates.
	 */
	low = 0;
	while (low < nt) {
		f1 = &t[low];
		for (i = low+1; i < nt; i++) {
			f2 = &t[i];
			if (f2->size != f1->size)
				break;
			dups.cmp++;
			if (cmp(path, f1->name, f2->name) == 0) {
				dups.rm++;
				sprintf(file, "%s/%s", path, f2->name);
				if (verbose > 1)
					printf("rm %s\n", file);
				unlink(file);
			}
		}
		low = i;
	}
}

/*
 * Same as "/bin/cmp -s dir/a dir/b".
 */
cmp(dir, a, b)
	char *dir, *a, *b;
{
	static struct {
		int fd;
		int n;
		char *b;
	} f[2];
	int i;

	if (f[0].b == 0)
		for (i = 0; i < 2; i++)	/* allocate I/O buffers */
			if ((f[i].b = valloc(IOSIZE)) == 0)
				Perror("valloc", " "), exit(1);
	sprintf(f[0].b, "%s/%s", dir, a);
	if ((f[0].fd = open(f[0].b, 0)) < 0)
		return -1;
	sprintf(f[0].b, "%s/%s", dir, b);
	if ((f[1].fd = open(f[0].b, 0)) < 0) {
		close(f[0].fd);
		return -1;
	}
        for (;;) {
		for (i = 0; i < 2; i++)
			f[i].n = read(f[i].fd, f[i].b, IOSIZE);
		if (f[0].n == f[1].n && f[0].n > 0 &&
		    memcmp(f[0].b, f[1].b, f[0].n) == 0)
			continue;
		close(f[0].fd);
		close(f[1].fd);
		return f[0].n == f[1].n && f[0].n == 0 ? 0 : 1;
        }
}
